/**
 * Copyright 2019-2023 Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of DesQ project (https://gitlab.com/desq/)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include "Global.hpp"
#include "PagerWidget.hpp"
#include "dbustypes.hpp"

#include <wayqt/WayQtUtils.hpp>
#include <wayqt/Registry.hpp>
#include <wayqt/DesQShell.hpp>

WorkspaceWidget::WorkspaceWidget( WorkSpace ws, int number, QWidget *parent ) : QWidget( parent ) {
    mCurWS  = ws;
    mNumber = number;
    setFixedSize( QSize( 32, 24 ) );

    setMouseTracking( true );
}


WorkSpace WorkspaceWidget::workSpace() {
    return mCurWS;
}


void WorkspaceWidget::highlight( bool yes ) {
    mHighlight = yes;
    repaint();
}


void WorkspaceWidget::mousePressEvent( QMouseEvent *mEvent ) {
    if ( mEvent->button() == Qt::LeftButton ) {
        mPressed = true;
    }

    mEvent->accept();
}


void WorkspaceWidget::mouseReleaseEvent( QMouseEvent *mEvent ) {
    if ( (mEvent->button() == Qt::LeftButton) and mPressed ) {
        /**
         * If we're highlighted, then this is the current workspace.
         * We toggle "ShowDesktop"
         */
        if ( mHighlight ) {
            output->call( "ShowDesktop" );
            emit showDesktop( mCurWS );
        }

        else {
            emit switchWorkSpace( mCurWS );
        }
    }

    repaint();
    mEvent->accept();
}


// void WorkspaceWidget::enterEvent( QEvent *event ) {
//     // emit showWorkSpaceInfo( mCurWS );
//
//     QWidget::enterEvent( event );
// }


void WorkspaceWidget::paintEvent( QPaintEvent *pEvent ) {
    QPainter painter( this );

    painter.setRenderHint( QPainter::Antialiasing );

    painter.save();

    if ( mHighlight ) {
        QColor clr( palette().color( QPalette::Highlight ) );
        clr = clr.darker();
        painter.setPen( QPen( clr, 1.0 ) );
        clr = clr.lighter();
        clr.setAlphaF( 0.5 );
        painter.setBrush( clr );
    }

    else {
        QColor clr( palette().color( QPalette::Window ) );
        painter.setPen( QPen( clr, 1.0 ) );
        clr.setAlphaF( 0.5 );
        painter.setBrush( clr );
    }

    painter.drawRoundedRect( QRectF( rect() ).adjusted( 0.5, 0.5, -0.5, -0.5 ), 3, 3 );
    painter.restore();

    painter.save();

    if ( mHighlight ) {
        QFont f( font() );
        f.setWeight( QFont::Bold );
        painter.setFont( f );
    }

    painter.drawText( rect(), Qt::AlignCenter, QString::number( mNumber ) );
    painter.restore();

    painter.end();

    pEvent->accept();
}


PagerPluginWidget::PagerPluginWidget( QWidget *parent ) : DesQ::Panel::PluginWidget( parent ) {
    qRegisterMetaType<QUIntList>( "QUIntList" );
    qDBusRegisterMetaType<QUIntList>();

    qRegisterMetaType<WorkSpace>( "WorkSpace" );
    qDBusRegisterMetaType<WorkSpace>();

    qRegisterMetaType<WorkSpaces>( "WorkSpaces" );
    qDBusRegisterMetaType<WorkSpaces>();

    qRegisterMetaType<WorkSpaceGrid>( "WorkSpaceGrid" );
    qDBusRegisterMetaType<WorkSpaceGrid>();

    QDBusConnection::sessionBus().connect(
        "org.DesQ.Wayfire",
        "/org/DesQ/Wayfire",
        "wayland.compositor",
        "ViewAdded",
        "u",
        this,
        SLOT( updatePopup( uint ) )
    );

    wsLyt = new QGridLayout();
    setLayout( wsLyt );

    connect(
        this, &DesQ::Panel::PluginWidget::panelScreenChanged, [ = ] ( QScreen *scrn ) {
            WQt::DesQShell *shell = nullptr;

            int time = 0;
            do {
                /** Sleep for 10 ms */
                QThread::msleep( 10 );
                time += 10;

                /** Process outstanding queue */
                qApp->processEvents();

                /** Attempt to retrieve DesQShell interface */
                shell = wlRegistry->desqShell();

                if ( time > 5000 ) {
                    qCritical() << "Unable to bind to DesQ Shell interface. Please ensure that the compositor supports it.";
                    break;
                }
            } while ( shell == nullptr );

            if ( shell ) {
                mDesQOp = shell->getOutput( WQt::Utils::wlOutputFromQScreen( scrn ) );

                if ( mDesQOp ) {
                    /** Store the incoming workspace grid in pending state */
                    connect(
                        mDesQOp, &WQt::DesQOutput::workspaceExtentsChanged, [ = ] ( int rows, int cols ) {
                            pendingGrid.rows    = rows;
                            pendingGrid.columns = cols;
                        }
                    );

                    /** Store the incoming workspace grid in pending state */
                    connect(
                        mDesQOp, &WQt::DesQOutput::activeWorkspaceChanged, [ = ] ( int row, int col ) {
                            pendingWS.row    = row;
                            pendingWS.column = col;
                        }
                    );

                    /** Commit our pending states */
                    connect(
                        mDesQOp, &WQt::DesQOutput::done, [ = ] () mutable {
                            if ( pendingGrid.rows and pendingGrid.columns ) {
                                populateLayout( pendingGrid );
                                pendingGrid.rows    = 0;
                                pendingGrid.columns = 0;
                            }

                            if ( pendingWS.row >= 0 and pendingWS.column >= 0 ) {
                                highlightWorkspace( pendingWS );
                                pendingWS.row    = -1;
                                pendingWS.column = -1;
                            }
                        }
                    );

                    mDesQOp->setup();
                }
            }
        }
    );

    /** So that we know exactly where the mouse is */
    setMouseTracking( true );

    popup = new QLabel();
    popup->setStyleSheet( "padding: 5px; border: 1px solid white; border-radius: 3px; background-color: rgba(0, 0, 0, 180);" );
    popup->setWindowFlags( Qt::Widget | Qt::FramelessWindowHint | Qt::BypassWindowManagerHint );
}


PagerPluginWidget::~PagerPluginWidget() {
    qDeleteAll( workspaceList );

    delete wsLyt;
    delete popup;
}


void PagerPluginWidget::populateLayout( WorkSpaceGrid grid ) {
    /* Delete the previously added WorkSpace widgets */
    if ( wsLyt->count() ) {
        qDeleteAll( wsLyt->findChildren<WorkspaceWidget *>() );
    }

    if ( workspaceList.count() ) {
        qDeleteAll( workspaceList );
    }

    workspaceList.clear();

    for ( int r = 0; r < grid.rows; r++ ) {
        for ( int c = 0; c < grid.columns; c++ ) {
            int             wsCount = r * grid.columns + c + 1;
            WorkspaceWidget *wsw    = new WorkspaceWidget( { r, c }, wsCount, this );
            wsLyt->addWidget( wsw, r, c );
            workspaceList << wsw;

            connect( wsw, &WorkspaceWidget::showWorkSpaceInfo, this, &PagerPluginWidget::showWorkSpaceInfo );
            connect(
                wsw, &WorkspaceWidget::switchWorkSpace, [ = ] ( WorkSpace ws ) {
                    mDesQOp->setActiveWorkspace( ws.row, ws.column );
                }
            );

            connect( wsw, &WorkspaceWidget::showWorkSpaceInfo, this, &PagerPluginWidget::showWorkSpaceInfo );
            connect(
                wsw, &WorkspaceWidget::hideTooltip, [ = ]() {
                    if ( popup->isVisible() ) {
                        popup->close();
                    }
                }
            );
        }
    }
}


void PagerPluginWidget::highlightWorkspace( WorkSpace ws ) {
    for ( WorkspaceWidget *wsw: workspaceList ) {
        if ( wsw->workSpace() == ws ) {
            wsw->highlight( true );
        }

        else {
            wsw->highlight( false );
        }
    }
}


void PagerPluginWidget::updatePopup( uint ) {
    if ( popup and popup->isVisible() ) {
        QPoint          cursor = mapFromGlobal( QCursor::pos() );
        WorkspaceWidget *ww    = qobject_cast<WorkspaceWidget *>( childAt( cursor ) );

        if ( ww ) {
            WorkSpace ws = ww->workSpace();
            showWorkSpaceInfo( ws );
        }
    }
}


void PagerPluginWidget::showWorkSpaceInfo( WorkSpace ws ) {
    QVariant wsVar;

    wsVar.setValue( ws );
    QDBusReply<QUIntList> reply = workspaces->call( "QueryWorkspaceViews", wsVar );

    if ( reply.isValid() ) {
        if ( reply.value().count() ) {
            QString info = QString( "Workspace (%1, %2)" ).arg( ws.row + 1 ).arg( ws.column + 1 );
            for ( uint vId: reply.value() ) {
                QDBusReply<QString> title = views->call( "QueryViewTitle", vId );

                if ( title.isValid() ) {
                    info += "<br>&nbsp;&nbsp;&nbsp;&nbsp;- " + title.value();
                }
            }

            popup->setText( info );

            if ( not popup->isVisible() ) {
                emit showTooltip( popup );
            }

            return;
        }
    }

    if ( popup->isVisible() ) {
        close();
    }
}
